use super::{archive, package, Handle};
use crate::db;

use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    sync::{atomic::AtomicU32, Arc, Mutex, RwLock},
};

use sea_orm::{
    ActiveModelTrait, ColumnTrait, Condition, ConnectionTrait, DbConn, EntityTrait, JoinType,
    ModelTrait, NotSet, QueryFilter, QuerySelect, Related, RelationTrait, Set, TransactionTrait,
};
use tokio::{
    runtime,
    sync::mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
};

pub enum RepoCommand {
    ParsePkg(i32, PathBuf),
}

pub struct RepoSharedState {
    repos_dir: PathBuf,
    conn: DbConn,
    rx: Mutex<UnboundedReceiver<RepoCommand>>,
    tx: UnboundedSender<RepoCommand>,
    repos: RwLock<HashMap<i32, (AtomicU32, Arc<Mutex<()>>)>>,
}

impl RepoSharedState {
    pub fn new(
        repos_dir: impl AsRef<Path>,
        conn: DbConn,
        repos: HashMap<i32, (AtomicU32, Arc<Mutex<()>>)>,
    ) -> Self {
        let (tx, rx) = unbounded_channel();

        Self {
            repos_dir: repos_dir.as_ref().to_path_buf(),
            conn,
            rx: Mutex::new(rx),
            tx,
            repos: RwLock::new(repos),
        }
    }
}

/// The actor is responsible for mutating the repositories. They receive their commands their
/// messages and process these commands in both a synchronous and asynchronous way.
pub struct RepoActor {
    rt: runtime::Handle,
    state: Arc<RepoSharedState>,
}

impl RepoActor {
    pub fn new(rt: runtime::Handle, state: Arc<RepoSharedState>) -> Self {
        Self {
            rt,
            state: Arc::clone(&state),
        }
    }

    /// Run the main actor loop
    pub fn run(self) {
        while let Some(msg) = {
            let mut rx = self.state.rx.lock().unwrap();
            rx.blocking_recv()
        } {
            match msg {
                RepoCommand::ParsePkg(repo, path) => {
                    let _ = self.parse_pkg(repo, path);
                }
            }
        }
    }

    fn parse_pkg(&self, repo: i32, path: PathBuf) -> crate::Result<()> {
        let pkg = package::Package::open(&path)?;
        let pkg = self
            .rt
            .block_on(db::query::package::insert(&self.state.conn, repo, pkg))?;
        let dest_path = self
            .state
            .repos_dir
            .join(repo.to_string())
            .join(pkg.id.to_string());
        std::fs::rename(path, dest_path)?;

        tracing::info!(
            "Added '{}-{}-{}' to repository {}",
            pkg.name,
            pkg.version,
            pkg.arch,
            repo,
        );

        Ok(())
    }
}
